# Copyright (c) 2016 QNAP Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver for QNAP Storage.
This driver supports QNAP Storage for iSCSI.
"""
import base64
import functools
import re
import six
import ssl
import time

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

from cinder import db
from cinder import exception
from cinder.i18n import _, _LE, _LI
from cinder.volume.drivers.san import san
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
from oslo_utils import units
from six.moves import http_client
from six.moves import urllib

DRIVER_VERSION = '1.0.0'

LOG = logging.getLogger(__name__)

qnap_opts = [
    cfg.StrOpt('qnap_server_port',
               default='8080',
               help='Management Port for QNAP storage'),
    cfg.StrOpt('qnap_storage_protocol',
               choices=['iscsi', 'nfs'],
               default='iscsi',
               help='Communication protocol to access QNAP storage'),
    cfg.StrOpt('qnap_poolname',
               default='pool1',
               help='The pool name in the QNAP Storage'),
]

CONF = cfg.CONF
CONF.register_opts(qnap_opts)


class QnapISCSIDriver(san.SanISCSIDriver):
    """OpenStack driver to enable QNAP Storage.

    Version history:
        1.0.0 - Initial driver (Only iSCSI)
    """

    def __init__(self, *args, **kwargs):
        super(QnapISCSIDriver, self).__init__(*args, **kwargs)
        self.APIExecutor = None
        self.group_stats = {}
        self.configuration.append_config_values(qnap_opts)

    def _check_config(self):
        """Ensure that the flags we care about are set."""
        required_config = ['san_ip', 'qnap_server_port', 'iscsi_ip_address',
                           'san_login', 'san_password']
        for attr in required_config:
            if not getattr(self.configuration, attr, None):
                raise exception.InvalidInput(reason=_('%s is not set.') %
                                             attr)

    def do_setup(self, context):
        """Setup the QNAP Cinder volume driver."""
        self._check_config()
        self.ctxt = context
        LOG.debug(_LI('test record context: %s'), context)

        # Setup API Executor
        try:
            self.APIExecutor = self.creatAPIExecutor()
        except Exception:
            LOG.error(_LE('Failed to create HTTP client. '
                          'Check ip, port, username, password'
                          ' and make sure the array version is compatible'))
            raise

    def creatAPIExecutor(self):
        self.APIExecutor = QnapAPIExecutor(
            username=self.configuration.san_login,
            password=self.configuration.san_password,
            ip=self.configuration.san_ip,
            port=self.configuration.qnap_server_port)

        nasModelName, isStorageV2, fwVersion, hostname = (
            self.APIExecutor.getBasicInfo(self.configuration.san_ip))

        pattern = re.compile(r"^([A-Z]+)-?\d+\S*$")
        matches = pattern.match(nasModelName)

        if not matches:
            return None
        modelType = matches.group(1)

        tsModelTypes = [
            "TS", "SS", "IS", "TVS", "TDS", "TBS"
        ]
        esModelTypes = [
            "ES"
        ]
        if modelType in tsModelTypes:
            if fwVersion.startswith("4.2"):
                LOG.debug(_LI('Create TS API Executor'))
                # modify the pool name to poole index
                pattern = re.compile(r"^(\d+)+|^Storage Pool (\d+)+")
                matches = pattern.match(self.configuration.qnap_poolname)
                LOG.debug(_LI('matches.group(1): %s'), matches.group(1))
                LOG.debug(_LI('matches.group(2): %s'), matches.group(2))
                if matches.group(1):
                    self.configuration.qnap_poolname = matches.group(1)
                else:
                    self.configuration.qnap_poolname = matches.group(2)

                return (QnapAPIExecutorTS(
                    username=self.configuration.san_login,
                    password=self.configuration.san_password,
                    ip=self.configuration.san_ip,
                    port=self.configuration.qnap_server_port))
        elif modelType in esModelTypes:
            if fwVersion.startswith("1.1."):
                LOG.debug(_LI('Create TS API Executor'))
                return (QnapAPIExecutor(
                    username=self.configuration.san_login,
                    password=self.configuration.san_password,
                    ip=self.configuration.san_ip,
                    port=self.configuration.qnap_server_port))

        return None

    def create_volume(self, volume):
        """Create a new volume."""
        LOG.debug(_LI('in create_volume'))
        reserve = self.configuration.san_thin_provision
        LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])

        if volume['display_name']:
            if len(volume['display_name']) > 31:
                LOG.debug(_LI('volume[display_name] is too long...'))
                LOG.debug(_LI('volume[:14]: %s'), volume['display_name'][:14])
                LOG.debug(_LI('volume[-14:]: %s'),
                          volume['display_name'][-14:])
                volume['display_name'] = (
                    volume['display_name'][:14] + '...' +
                    volume['display_name'][-14:])
                LOG.debug(_LI('volume[display_name]: %s'),
                          volume['display_name'])
        else:
            LOG.debug(_LI('volume[display_name] is None...'))
            volume['display_name'] = ('QCD-' + volume['name'][7:].replace(
                '-', ''))[0:31]

        self.APIExecutor._execute_create_vol(volume,
                                             self.configuration.qnap_poolname,
                                             reserve)

        createdLun = self.APIExecutor.getLunInfo(
            LUNName=volume['display_name'])
        LUNNAA = createdLun.find('LUNNAA').text

        timeIntervalM = 3
        maxWaitSec = 600
        tryTimes = 0
        while(tryTimes < maxWaitSec and not LUNNAA):
            time.sleep(timeIntervalM)
            createdLun = self.APIExecutor.getLunInfo(
                LUNName=volume['display_name'])
            LUNNAA = createdLun.find('LUNNAA').text
            tryTimes = tryTimes + 3

        LOG.debug(_LI('LUNNAA: %s'), LUNNAA)
        _metadata = {'LUNNAA': LUNNAA}
        db.volume_metadata_update(self.ctxt,
                                  volume['id'],
                                  _metadata,
                                  False)

    def delete_volume(self, volume):
        """Delete the specified volume."""
        vol_metadata = self.get_volume_metadata(volume['id'])
        LOG.debug(_LI('vol_metadata: %s'), vol_metadata)
        LOG.debug(_LI('len(vol_metadata): %d'), len(vol_metadata))
        LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])
        if len(vol_metadata) != 0 and not volume['display_name']:
            lun_naa = vol_metadata['LUNNAA']
            LOG.debug(_LI('LUNNAA: %s'), lun_naa)
        elif volume['display_name']:
            try:
                if len(volume['display_name']) > 31:
                    LOG.debug(_LI('volume[display_name] is too long...'))
                    LOG.debug(_LI('volume[:14]: %s'),
                              volume['display_name'][:14])
                    LOG.debug(_LI('volume[-14:]: %s'),
                              volume['display_name'][-14:])
                    volume['display_name'] = (
                        volume['display_name'][:14] +
                        '...' +
                        volume['display_name'][-14:])
                    LOG.debug(_LI('volume[display_name]: %s'),
                              volume['display_name'])

                createdLun = self.APIExecutor.getLunInfo(
                    LUNName=volume['display_name'])
                lun_naa = createdLun.find('LUNNAA').text
                LOG.debug(_LI('LUNNAA: %s'), lun_naa)
                _metadata = {'LUNNAA': lun_naa}
                db.volume_metadata_update(self.ctxt,
                                          volume['id'],
                                          _metadata,
                                          False)
            except Exception:
                LOG.debug(_LI('Volume is invalid'))
        else:
            # no volume exist
            return

        try:
            delLun = self.APIExecutor.getLunInfo(LUNNAA=lun_naa)
            LUNIndex = delLun.find('LUNIndex').text
            LOG.debug(_LI('LUNIndex: %s'), LUNIndex)
            self.APIExecutor.delete_vol(LUNIndex)
        except Exception:
            LOG.debug(_LI('Volume is invalid'))

    def get_volume_metadata(self, volume_id):
        return self.db.volume_metadata_get(self.ctxt, volume_id)

    def _extend_lun(self, volume):
        selectedLun = self.APIExecutor.getLunInfo(
            LUNName=volume['display_name'])
        LUNIndex = selectedLun.find('LUNIndex').text
        LOG.debug(_LI('LUNIndex: %s'), LUNIndex)
        LUNName = selectedLun.find('LUNName').text
        LOG.debug(_LI('LUNName: %s'), LUNName)
        LUNThinAllocate = selectedLun.find('LUNThinAllocate').text
        LOG.debug(_LI('LUNThinAllocate: %s'), LUNThinAllocate)
        LUNPath = ''
        if selectedLun.find('LUNPath') is not None:
            LUNPath = selectedLun.find('LUNPath').text
            LOG.debug(_LI('LUNPath: %s'), LUNPath)
        LUNStatus = selectedLun.find('LUNStatus').text
        LOG.debug(_LI('LUNStatus: %s'), LUNStatus)

        lun = {'LUNName': LUNName,
               'LUNCapacity': volume['size'],
               'LUNIndex': LUNIndex,
               'LUNThinAllocate': LUNThinAllocate,
               'LUNPath': LUNPath,
               'LUNStatus': LUNStatus}
        self.APIExecutor.editLun(lun)

    def create_cloned_volume(self, volume, src_vref):
        """Create a clone of the specified volume."""
        LOG.debug(_LI('Entering create_cloned_volume...'))

        # Below is to clone a volume from a snapshot in the snapshot manager
        srcLun = self.APIExecutor.getLunInfo(
            LUNName=src_vref['display_name'])
        LUNIndex = srcLun.find('LUNIndex').text
        LOG.debug(_LI('LUNIndex: %s'), LUNIndex)

        if len(volume['display_name']) > 31:
            LOG.debug(_LI('volume[display_name] is too long...'))
            LOG.debug(_LI('volume[:14]: %s'), volume['display_name'][:14])
            LOG.debug(_LI('volume[-14:]: %s'), volume['display_name'][-14:])
            volume['display_name'] = (
                volume['display_name'][:14] +
                '...' +
                volume['display_name'][-14:])
            LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])

        ret = self.APIExecutor.createSnapshot(LUNIndex, volume['display_name'])
        root = ET.fromstring(ret['data'])
        snapshot_id = root.find('SnapshotInfo').find('snapshot_id').text
        LOG.debug(_LI('snapshot_id: %s'), snapshot_id)

        ret = self.APIExecutor.cloneSnapshot(snapshot_id,
                                             volume['display_name'])

        if (volume['size'] > src_vref['size']):
            self._extend_lun(volume)

    def create_export(self, context, volume, connector=None):
        """Driver entry point to get the export info for a new volume."""
        LOG.debug(_LI('In create_export'))
        model_update = volume['provider_location']
        LOG.debug(_LI('model_update: %s'), model_update)
        LOG.debug(_LI('Create export done from Volume %(volume_id)s.'), {
            'volume_id': volume['id']})
        return {'provider_location': model_update}

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume."""

    def create_snapshot(self, snapshot):
        """Create a snapshot."""
        LOG.debug(_LI('Entering create_snapshot snapshot=%(snap)s'),
                  {'snap': snapshot.volume.display_name})

        # Below is to create snapshot in the snapshot manager
        vol_metadata = self.get_volume_metadata(snapshot.volume.id)
        if len(vol_metadata) != 0:
            lun_naa = vol_metadata['LUNNAA']
            LOG.debug(_LI('LUNNAA: %s'), lun_naa)
        else:
            if len(snapshot.volume.display_name) > 31:
                LOG.debug(_LI('volume[display_name] is too long...'))
                LOG.debug(_LI('volume[:14]: %s'),
                          snapshot.volume.display_name[:14])
                LOG.debug(_LI('volume[-14:]: %s'),
                          snapshot.volume.display_name[-14:])
                snapshot.volume.display_name = (
                    snapshot.volume.display_name[:14] +
                    '...' +
                    snapshot.volume.display_name[-14:])
                LOG.debug(_LI('[display_name]: %s'),
                          snapshot.volume.display_name)

            createdLun = self.APIExecutor.getLunInfo(
                LUNName=snapshot.volume.display_name)
            lun_naa = createdLun.find('LUNNAA').text
            LOG.debug(_LI('LUNNAA: %s'), lun_naa)

        _metadata = {'LUNNAA': lun_naa}
        db.volume_metadata_update(self.ctxt,
                                  snapshot.volume.id,
                                  _metadata,
                                  False)

        try:
            srcLun = self.APIExecutor.getLunInfo(LUNNAA=lun_naa)
            LUNIndex = srcLun.find('LUNIndex').text
            LOG.debug(_LI('LUNIndex: %s'), LUNIndex)
            mod_snapshot = snapshot.volume.display_name.replace(" ", "-")
            LOG.debug(_LI('snapshot[display_name]2: %s'), mod_snapshot)

            if len(mod_snapshot) > 31:
                LOG.debug(_LI('mod_snapshot is too long...'))
                LOG.debug(_LI('mod_snapshot[:14]: %s'), mod_snapshot[:14])
                LOG.debug(_LI('mod_snapshot[-14:]: %s'), mod_snapshot[-14:])
                mod_snapshot = (
                    mod_snapshot[:14] +
                    '...' +
                    mod_snapshot[-14:])
                LOG.debug(_LI('mod_snapshot: %s'), mod_snapshot)

            ret = self.APIExecutor.createSnapshot(LUNIndex, mod_snapshot)
            root = ET.fromstring(ret['data'])
            # result of TS NAS will be the index of snapsthot
            # result of ES NAS will be the status of creating
            if root.find('result').text >= '0':
                snapshot_id = (root.find('SnapshotInfo')
                                   .find('snapshot_id').text)
                LOG.debug(_LI('snapshot_id: %s'), snapshot_id)

                _metadata = {'snapshot_id': snapshot_id}
                db.snapshot_metadata_update(self.ctxt,
                                            snapshot['id'],
                                            _metadata,
                                            False)
        except exception.QnapAPIExpiredException:
            raise exception.QnapAPIExpiredException(_('session id expired'))

    def delete_snapshot(self, snapshot):
        """Delete a snapshot."""
        LOG.debug(_LI('Entering delete_snapshot snapshot name=%(name)s '
                  'volume_name=%(volname)s'),
                  {'name': snapshot.name, 'volname': snapshot.volume_name})

        # Below is to delete snapshot in the snapshot manager
        snap_metadata = self.get_snapshot_metadata(snapshot['id'])

        try:
            snapshot_id = snap_metadata['snapshot_id']
            LOG.debug(_LI('snapshot_id: %s'), snapshot_id)
            self.APIExecutor.deleteSnapshot(snapshot_id)
        except Exception:
            LOG.debug(_LI('Metadata of the snapshot is invalid'))

    def get_snapshot_metadata(self, snapshot_id):
        return self.db.snapshot_metadata_get(self.ctxt, snapshot_id)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""
        LOG.debug(_LI('in create_volume_from_snapshot'))
        LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])
        LOG.debug(_LI('snapshot[id]: %s'), snapshot['id'])
        LOG.debug(_LI('volume: %s'), volume)
        LOG.debug(_LI('snapshot: %s'), snapshot)
        # Below is to clone a volume from a snapshot in the snapshot manager
        snap_metadata = self.get_snapshot_metadata(snapshot['id'])
        try:
            snapshot_id = snap_metadata['snapshot_id']
            if volume['display_name']:
                if len(volume['display_name']) > 31:
                    LOG.debug(_LI('volume[display_name] is too long...'))
                    LOG.debug(_LI('volume[:14]: %s'),
                              volume['display_name'][:14])
                    LOG.debug(_LI('volume[-14:]: %s'),
                              volume['display_name'][-14:])
                    volume['display_name'] = (
                        volume['display_name'][:14] +
                        '...' +
                        volume['display_name'][-14:])
                    LOG.debug(_LI('volume[display_name]: %s'),
                              volume['display_name'])
            else:
                volume['display_name'] = ('QCD-' + volume['name'][7:].replace(
                    '-', ''))[0:31]

            self.APIExecutor.cloneSnapshot(snapshot_id, volume['display_name'])
            createdLun = self.APIExecutor.getLunInfo(
                LUNName=volume['display_name'])

        except Exception:
            LOG.debug(_LI('Metadata of the snapshot is invalid'))
            msg = 'Metadata of the snapshot is invalid'
            raise exception.QnapDriverException(message=msg)

        LUNNAA = createdLun.find('LUNNAA').text
        LOG.debug(_LI('LUNNAA: %s'), LUNNAA)

        _metadata = {'LUNNAA': LUNNAA}
        db.volume_metadata_update(self.ctxt,
                                  volume['id'],
                                  _metadata,
                                  False)

        if (volume['size'] > snapshot['volume_size']):
            self._extend_lun(volume)

    def get_volume_stats(self, refresh=False):
        """Get volume stats. This is more of getting group stats."""
        LOG.debug(_LI('in get_volume_stats'))

        if refresh:
            backend_name = (self.configuration.safe_get(
                            'volume_backend_name') or
                            self.__class__.__name__)
            LOG.debug(_LI('backend_name=%(backend_name)s'),
                      {'backend_name': backend_name})

            selectedPool = self.APIExecutor.getSpecificPoolInfo(self.
                                                                configuration.
                                                                qnap_poolname)
            capacity_bytes = int(selectedPool.find('capacity_bytes').text)
            LOG.debug(_LI('capacity_bytes: %s GB'), capacity_bytes / units.Gi)
            freesize_bytes = int(selectedPool.find('freesize_bytes').text)
            LOG.debug(_LI('freesize_bytes: %s GB'), freesize_bytes / units.Gi)

            self.group_stats = {'volume_backend_name': backend_name,
                                'vendor_name': 'QNAP',
                                'driver_version': DRIVER_VERSION,
                                'storage_protocol': 'iSCSI',
                                'total_capacity_gb': capacity_bytes / units.Gi,
                                'free_capacity_gb': freesize_bytes / units.Gi,
                                'reserved_percentage': 0,
                                'QoS_support': False}
            return self.group_stats

    def extend_volume(self, volume, new_size):
        """Extend an existing volume."""
        LOG.debug(_LI('Entering extend_volume volume=%(vol)s '
                  'new_size=%(size)s'),
                  {'vol': volume['display_name'], 'size': new_size})

        if len(volume['display_name']) > 31:
            LOG.debug(_LI('volume[display_name] is too long...'))
            LOG.debug(_LI('volume[:14]: %s'), volume['display_name'][:14])
            LOG.debug(_LI('volume[-14:]: %s'), volume['display_name'][-14:])
            volume['display_name'] = (
                volume['display_name'][:14] +
                '...' +
                volume['display_name'][-14:])
            LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])

        volume['size'] = new_size
        self._extend_lun(volume)

    def initialize_connection(self, volume, connector):
        """Create a target with initiator iqn to attach a volume."""
        LOG.debug(_LI('Entering initialize_connection volume=%(vol)s'
                  ' connector=%(conn)s location=%(loc)s'),
                  {'vol': volume,
                   'conn': connector,
                   'loc': volume['provider_location']})

        selectedPool = self.APIExecutor.getPoolInfo(
            poolID=self.configuration.qnap_poolname)

        controllerName = ''
        LOG.debug(_LI('pool_owner: %s'), selectedPool.find('pool_owner'))
        if selectedPool.find('pool_owner') is not None:
            controllerName = selectedPool.find('pool_owner').text
        LOG.debug(_LI('controllerName: %s'), controllerName)

        ret = self.APIExecutor.getIscsiPortalInfo()
        root = ET.fromstring(ret['data'])
        iscsiPort = root.find('iSCSIPortal').find('servicePort').text
        LOG.debug(_LI('iscsiPort: %s'), iscsiPort)
        targetIQNPrefix = root.find('iSCSIPortal').find('targetIQNPrefix').text
        LOG.debug(_LI('targetIQNPrefix: %s'), targetIQNPrefix)
        targetIQNPostfix = (root.find('iSCSIPortal').
                            find('targetIQNPostfix').text)
        LOG.debug(_LI('targetIQNPostfix: %s'), targetIQNPostfix)

        LOG.debug(_LI('volume[name]: %s'), volume['name'])
        LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])

        if volume['display_name']:
            if len(volume['display_name']) > 31:
                LOG.debug(_LI('volume[display_name] is too long...'))
                LOG.debug(_LI('volume[:14]: %s'), volume['display_name'][:14])
                LOG.debug(_LI('volume[-14:]: %s'),
                          volume['display_name'][-14:])
                volume['display_name'] = (
                    volume['display_name'][:14] +
                    '...' +
                    volume['display_name'][-14:])
                LOG.debug(_LI('volume[display_name]: %s'),
                          volume['display_name'])
            LUNIndex = self.APIExecutor._getLunIDByLunname(
                LUNName=volume['display_name'])
        else:
            new_volname = ('QCD-' + volume['name'][7:].replace('-', ''))[0:31]
            LUNIndex = self.APIExecutor._getLunIDByLunname(LUNName=new_volname)
        LOG.debug(_LI('aaa LUNIndex: %s'), LUNIndex)

        ret = self.APIExecutor.getAllIscsiPortalSetting()
        root = ET.fromstring(ret['data'])

        targetIndex = ''
        targetIqn = ''
        # find the targets have acl with connector['initiator']
        targetWithInitiatorList = []
        targetACLTree = root.find('targetACL')
        targetACLList = targetACLTree.findall('row')
        for targetACL in targetACLList:
            targetInitInfoList = targetACL.findall('targetInitInfo')
            for targetInitInfo in targetInitInfoList:
                if(targetInitInfo.find('initiatorIQN').text ==
                   connector['initiator']):
                    targetWithInitiatorList.append(
                        targetACL.find('targetIndex').text)

        # find the target in targetWithInitiatorList with ready status
        targetTree = root.find('iSCSITargetList')
        targetList = targetTree.findall('targetInfo')
        for targetWithInitiator in targetWithInitiatorList:
            for target in targetList:
                if(targetWithInitiator == target.find('targetIndex').text):
                    if int(target.find('targetStatus').text) >= 0:
                        targetIndex = targetWithInitiator
                        targetIqn = target.find('targetIQN').text

        # create a new target if no target has ACL connector['initiator']
        LOG.debug(_LI('exist targetIndex: %s'), targetIndex)
        if not targetIndex:
            targetName = "cinder-{0}".format(timeutils.
                                             utcnow().
                                             strftime('%Y%m%d%H%M%S%f'))
            LOG.debug(_LI('target_name: %s'), targetName)
            ret = self.APIExecutor.createTarget(targetName)
            root = ET.fromstring(ret['data'])
            targetIndex = root.find('result').text
            LOG.debug(_LI('targetIndex: %s'), targetIndex)
            targetIqn = targetIQNPrefix + targetName + targetIQNPostfix
            LOG.debug(_LI('controllerName: %s'), controllerName)
            if controllerName:
                targetIqn = (targetIqn +
                             ('.0' if controllerName == 'SCA' else '.1'))
            LOG.debug(_LI('targetIqn: %s'), targetIqn)

            # TS NAS will have to remove default ACL
            defaultACL = targetIQNPrefix[:targetIQNPrefix.find(":") + 1]
            defaultACL = defaultACL + "all:iscsi.default.ffffff"
            LOG.debug(_LI('defaultACL: %s'), defaultACL)
            self.APIExecutor.removeTargetInit(targetIqn, defaultACL)

            self.APIExecutor.addTargetInit(targetIqn, connector['initiator'])

        self.APIExecutor.mapLun(LUNIndex, targetIndex)

        properties = {}
        properties['target_discovered'] = True
        properties['target_portal'] = (self.configuration.iscsi_ip_address +
                                       ':' + iscsiPort)

        properties['target_iqn'] = targetIqn
        LOG.debug(_LI('properties[target_iqn]: %s'), properties['target_iqn'])
        vol_metadata = self.get_volume_metadata(volume['id'])
        lun_naa = vol_metadata['LUNNAA']
        LOG.debug(_LI('LUNNAA: %s'), lun_naa)
        mappedLun = self.APIExecutor.getLunInfo(LUNNAA=lun_naa)
        target_lun_id = int(mappedLun.find('LUNTargetList').find(
            'row').find('LUNNumber').text)
        LOG.debug(_LI('target_lun_id: %s'), target_lun_id)
        properties['target_lun'] = target_lun_id
        properties['volume_id'] = volume['id']  # used by xen currently
        properties['access_mode'] = 'rw'

        """Below are settings for multipath"""
        ethList = self.APIExecutor.getAllEthernetIp(type='all')
        target_portals = []
        target_iqns = []
        for eth in ethList:
            if eth['IP'] is not None:
                if eth['ethdev'] != 'em0':
                    target_portals.append(eth['IP'] + ':' + iscsiPort)
                    if 'ethSlotid' in eth:
                        target_iqns.append(targetIqn[:-2] + '.' +
                                           eth['ethSlotid'])
                    else:
                        target_iqns.append(targetIqn)
        properties['target_portals'] = target_portals
        properties['target_iqns'] = target_iqns
        properties['target_luns'] = [target_lun_id] * len(target_portals)
        LOG.debug(_LI('properties: %s'), properties)
        """"""
        return {
            'driver_volume_type': 'iscsi',
            'data': properties,
        }

    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to unattach a volume from an instance."""
        LOG.debug(_LI('Entering terminate_connection volume=%(vol)s'
                  ' connector=%(conn)s location=%(loc)s.'),
                  {'vol': volume,
                   'conn': connector,
                   'loc': volume['provider_location']})

        LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])

        if volume['display_name']:
            if len(volume['display_name']) > 31:
                LOG.debug(_LI('volume[display_name] is too long...'))
                LOG.debug(_LI('volume[:14]: %s'), volume['display_name'][:14])
                LOG.debug(_LI('volume[-14:]: %s'),
                          volume['display_name'][-14:])
                volume['display_name'] = (
                    volume['display_name'][:14] +
                    '...' +
                    volume['display_name'][-14:])
                LOG.debug(_LI('volume[display_name]: %s'),
                          volume['display_name'])
        else:
            volume['display_name'] = ('QCD-' + volume['name'][7:].replace(
                '-', ''))[0:31]
            LOG.debug(_LI('volume[display_name]: %s'),
                      volume['display_name'])

        selectedLun = self.APIExecutor.getLunInfo(
            LUNName=volume['display_name'])

        LUNIndex = selectedLun.find('LUNIndex').text
        LOG.debug(_LI('LUNIndex: %s'), LUNIndex)
        targetIndex = (selectedLun.find('LUNTargetList')
                       .find('row').find('targetIndex').text)
        LOG.debug(_LI('targetIndex: %s'), targetIndex)

        self.APIExecutor.unmapLun(LUNIndex, targetIndex)

    def migrate_volume(self, ctxt, volume, host):
        """Migrate the volume to the specified host."""
        LOG.debug(_LI('migrate_volume volume id=%(volume_id)s host=%(host)s'),
                  {'volume_id': volume['id'], 'host': host['host']})
        LOG.debug(_LI('host: %s'), host)

        return True, {}


def _connection_checker(func):
    """Decorator to check session has expired or not."""
    @functools.wraps(func)
    def inner_connection_checker(self, *args, **kwargs):
        LOG.debug(_LI('in _connection_checker'))
        for attempts in range(5):
            try:
                return func(self, *args, **kwargs)
            except exception.QnapAPIExpiredException as e:
                if attempts < 1:
                    LOG.debug(_LI('Session might have expired.'
                              ' Trying to relogin'))
                    self.login()
                    continue
                else:
                    LOG.error(_LE('Re-throwing Exception %s'), e)
                    raise
    return inner_connection_checker


class QnapAPIExecutor(object):
    """Makes QNAP API calls for ES NAS."""

    def __init__(self, *args, **kwargs):
        self.sid = None
        self.username = kwargs['username']
        self.password = kwargs['password']
        self.ip = kwargs['ip']
        self.port, self.ssl = self.getNasPortAndSsl(self.ip)
        self.login()

    def getNasPortAndSsl(self, host):
        LOG.debug(_LI('in getNasPortAndSsl'))
        LOG.debug(_LI('host: %s'), host)
        connection = http_client.HTTPConnection(host, 8080)
        connection.request('GET', '/cgi-bin/authLogin.cgi')
        response = connection.getresponse()
        data = response.read()
        LOG.debug(_LI('response data: %s'), data)

        root = ET.fromstring(data)
        QWebSSLEnabled = root.find('wfmSSLEnabled').text
        # need to confirm if set the ssl in cinder.conf
        if QWebSSLEnabled == '1':
            port = root.find('wfmSSLPort').text
            ssl = True
            return port, ssl
        else:
            port = root.find('wfmPort').text
            ssl = False
            return port, ssl

    def getBasicInfo(self, host):
        LOG.debug(_LI('in getBasicInfo'))
        connection = http_client.HTTPConnection(host, 8080)
        connection.request('GET', '/cgi-bin/authLogin.cgi')
        response = connection.getresponse()
        data = response.read()
        LOG.debug(_LI('response data: %s'), data)

        root = ET.fromstring(data)

        nasModelName = root.find('model/displayModelName').text
        isStorageV2 = root.find('model/storage_v2').text
        fwVersion = root.find('firmware/version').text
        hostname = root.find('hostname').text

        return nasModelName, isStorageV2, fwVersion, hostname

    def _execute_and_get_response_details(self, host, url):
        """Will prepare response after executing an http request. """
        LOG.debug(_LI('port: %(port)s, ssl: %(ssl)s'),
                  {'port': self.port, 'ssl': self.ssl})

        res_details = {}

        # Prepare the connection
        if self.ssl:
            if hasattr(ssl, '_create_unverified_context'):
                context = ssl._create_unverified_context()
                connection = http_client.HTTPSConnection(host,
                                                         port=self.port,
                                                         context=context)
            else:
                connection = http_client.HTTPSConnection(host, port=self.port)
        else:
            connection = http_client.HTTPConnection(host, self.port)

        # Make the connection
        LOG.debug(_LI('url : %s'), url)
        connection.request('GET', url)
        # Extract the response as the connection was successful
        response = connection.getresponse()
        # Read the response
        data = response.read()
        LOG.debug(_LI('response data: %s'), data)
        # Extract http error msg if any
        error_details = None
        res_details['data'] = data
        res_details['error'] = error_details
        res_details['http_status'] = response.status

        connection.close()
        return res_details

    def _execute_login(self):
        params = {}
        params['user'] = self.username
        params['pwd'] = base64.b64encode(self.password.encode("utf-8"))
        params['serviceKey'] = '1'

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/authLogin.cgi?%s' % sanitized_params)
        LOG.debug(_LI('url: %s'), url)

        res_details = self._execute_and_get_response_details(self.ip, url)
        return res_details['data']

    def login(self):
        """Execute Https Login API."""
        response = self._execute_login()
        LOG.debug(_LI('Successful login by user %s'), self.username)
        root = ET.fromstring(response)
        self.sid = root.find('authSid').text
        LOG.debug(_LI('sid: %s'), self.sid)

    @_connection_checker
    def _execute_create_vol(self, volume, pool_name, reserve):
        params = {}
        params['func'] = 'add_lun'
        params['FileIO'] = 'no'
        if reserve:
            params['LUNThinAllocate'] = '1'
        else:
            params['LUNThinAllocate'] = '0'

        if len(volume['display_name']) > 31:
            LOG.debug(_LI('volume[display_name] is too long...'))
            LOG.debug(_LI('volume[:14]: %s'), volume['display_name'][:14])
            LOG.debug(_LI('volume[-14:]: %s'), volume['display_name'][-14:])
            volume['display_name'] = (
                volume['display_name'][:14] +
                '...' +
                volume['display_name'][-14:])
            LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])

        LOG.debug(_LI('volume[display_name]: %s'), volume['display_name'])
        params['LUNName'] = volume['display_name']
        params['LUNPath'] = volume['display_name']
        params['poolID'] = pool_name
        params['lv_ifssd'] = 'no'
        params['LUNCapacity'] = volume['size']
        params['lv_threshold'] = '80'
        params['sid'] = self.sid
        sanitized_params = {}
        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_lun_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def _getLunIDByLunname(self, **kwargs):
        """Execute getLunInfo API."""
        for key, value in six.iteritems(kwargs):
            LOG.debug(_LI('%(key)s = %(val)s'),
                      {'key': key, 'val': value})

        params = {}
        params['func'] = 'extra_get'
        params['lunList'] = '1'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_portal_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        lunList = root.find('iSCSILUNList')
        lunInfoTree = lunList.findall('LUNInfo')
        for lun in lunInfoTree:
            LOG.debug(_LI('Iterating lun name: %s'), lun.find('LUNName').text)
            LOG.debug(_LI('kwargs[LUNName]: %s'), kwargs['LUNName'])
            if kwargs['LUNName'] == lun.find('LUNName').text:
                LOG.debug(_LI('matched......'))
                return lun.find('LUNIndex').text

    @_connection_checker
    def delete_vol(self, vol_id, *args, **kwargs):
        """Execute deleteVol API."""
        LOG.debug(_LI('Deleting volume id %s '), vol_id)
        params = {}
        params['func'] = 'remove_lun'
        params['run_background'] = '1'
        params['ha_sync'] = '1'
        params['LUNIndex'] = vol_id
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_lun_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def getPoolInfo(self, **kwargs):
        """Execute deleteInitiatorGrp API."""
        for key, value in six.iteritems(kwargs):
            LOG.debug(_LI('%(key)s = %(val)s'),
                      {'key': key, 'val': value})

        params = {}
        params['store'] = 'poolList'
        params['func'] = 'extra_get'
        params['extra_pool_index'] = '1'
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/disk_manage.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        if 'poolIndex' in kwargs:
            LOG.debug(_LI('in poolIndex section...'))
            root = ET.fromstring(res_details['data'])
            poolList = root.find('Pool_Index')
            poolInfoTree = poolList.findall('row')
            for pool in poolInfoTree:
                LOG.debug(_LI('Iterating pool name: %(name)s, index: %(id)s'),
                          {'name': pool.find('poolID').text,
                          'id': pool.find('poolIndex').text})
                if kwargs['poolIndex'] == pool.find('poolIndex').text:
                    LOG.debug(_LI('poolIndex: %s'),
                              pool.find('poolIndex').text)
                    return pool
        elif 'poolID' in kwargs:
            LOG.debug(_LI('in poolID section...'))
            root = ET.fromstring(res_details['data'])
            poolList = root.find('Pool_Index')
            poolInfoTree = poolList.findall('row')
            for pool in poolInfoTree:
                LOG.debug(_LI('Iterating pool name: %(name)s, index: %(id)s'),
                          {'name': pool.find('poolID').text,
                          'id': pool.find('poolIndex').text})
                if kwargs['poolID'] == pool.find('poolID').text:
                    LOG.debug(_LI('poolID: %s'), pool.find('poolID').text)
                    return pool
        else:
            return res_details['data']

    @_connection_checker
    def getSpecificPoolInfo(self, poolID):
        """Execute deleteInitiatorGrp API."""
        params = {}
        params['store'] = 'poolInfo'
        params['func'] = 'extra_get'
        params['poolID'] = poolID
        params['Pool_Info'] = '1'
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/disk_manage.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)

        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        poolList = root.find('Pool_Index')
        poolInfoTree = poolList.findall('row')
        for pool in poolInfoTree:
            if poolID == pool.find('poolID').text:
                LOG.debug(_LI('poolID: %s'), pool.find('poolID').text)
                return pool

    @_connection_checker
    def createTarget(self, targetName):
        params = {}
        params['func'] = 'add_target'
        params['targetName'] = targetName
        params['targetAlias'] = targetName
        params['bTargetDataDigest'] = '0'
        params['bTargetHeaderDigest'] = '0'
        params['bTargetClusterEnable'] = '1'
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_target_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def addTargetInit(self, targetIqn, initIqn):
        LOG.debug(_LI('targetIqn = %(tgt)s, initIqn = %(init)s'),
                  {'tgt': targetIqn, 'init': initIqn})

        params = {}
        params['func'] = 'add_init'
        params['targetIQN'] = targetIqn
        params['initiatorIQN'] = initIqn
        params['initiatorAlias'] = initIqn
        params['bCHAPEnable'] = '0'
        params['CHAPUserName'] = ''
        params['CHAPPasswd'] = ''
        params['bMutualCHAPEnable'] = '0'
        params['mutualCHAPUserName'] = ''
        params['mutualCHAPPasswd'] = ''
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_target_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

    def removeTargetInit(self, targetIqn, initIqn):
        pass

    @_connection_checker
    def mapLun(self, LUNIndex, targetIndex):
        LOG.debug(_LI('LUNIndex: %(lun)s, targetIndex: %(tgt)s'),
                  {'lun': LUNIndex, 'tgt': targetIndex})
        params = {}
        params['func'] = 'add_lun'
        params['LUNIndex'] = LUNIndex
        params['targetIndex'] = targetIndex
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_target_setting.cgi?%s' % sanitized_params)
        LOG.debug(_LI('mapLun url: %s'), url)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

    @_connection_checker
    def unmapLun(self, LUNIndex, targetIndex):
        params = {}
        params['func'] = 'remove_lun'
        params['LUNIndex'] = LUNIndex
        params['targetIndex'] = targetIndex
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_target_setting.cgi?%s' % sanitized_params)
        LOG.debug(_LI('unmapLun url: %s'), url)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

    @_connection_checker
    def getIscsiPortalInfo(self):
        params = {}
        params['func'] = 'extra_get'
        params['iSCSI_portal'] = '1'
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_portal_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def getLunInfo(self, **kwargs):
        """Execute getLunInfo API."""
        for key, value in six.iteritems(kwargs):
            LOG.debug(_LI('%(key)s = %(val)s'),
                      {'key': key, 'val': value})

        params = {}
        params['func'] = 'extra_get'
        params['lunList'] = '1'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_portal_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        if (('LUNIndex' in kwargs) or ('LUNName' in kwargs) or
           ('LUNNAA' in kwargs)):

            lunList = root.find('iSCSILUNList')
            lunInfoTree = lunList.findall('LUNInfo')
            for lun in lunInfoTree:
                LOG.debug(_LI('Iterating lun name: %(name)s, index: %(id)s'),
                          {'name': lun.find('LUNName').text,
                          'id': lun.find('LUNIndex').text})
                if ('LUNIndex' in kwargs):
                    if (kwargs['LUNIndex'] == lun.find('LUNIndex').text):
                        LOG.debug(_LI('LUNIndex:%s'),
                                  lun.find('LUNIndex').text)
                        return lun
                elif ('LUNName' in kwargs):
                    if (kwargs['LUNName'] == lun.find('LUNName').text):
                        LOG.debug(_LI('LUNName:%s'), lun.find('LUNName').text)
                        return lun
                elif ('LUNNAA' in kwargs):
                    if (kwargs['LUNNAA'] == lun.find('LUNNAA').text):
                        LOG.debug(_LI('LUNNAA:%s'), lun.find('LUNNAA').text)
                        return lun
        else:
            return res_details['data']

    @_connection_checker
    def createSnapshot(self, lunID, snapshot_name):
        """Execute CGI to create snapshot from source lun NAA."""
        params = {}
        params['func'] = 'create_snapshot'
        params['lunID'] = lunID
        params['snapshot_name'] = snapshot_name
        params['expire_min'] = '0'
        params['vital'] = '1'
        params['snapshot_type'] = '0'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/snapshot.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def createVssSnapshot(self, vssID, srcLunNAA):
        """Execute CGI to create snapshot from source lun NAA."""
        params = {}
        params['func'] = 'vss_create_snapshot'
        params['vssID'] = vssID
        params['LUNNAA'] = srcLunNAA
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/snapshot.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def deleteSnapshot(self, snapshotID):
        """Execute CGI to create snapshot from source lun NAA."""
        params = {}
        params['func'] = 'del_snapshots'
        params['snapshotID'] = snapshotID
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/snapshot.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
        else:
            return res_details

    @_connection_checker
    def deleteVssSnapshot(self, vssID, snapLunNAA):
        """Execute CGI to create snapshot from source lun NAA."""
        params = {}
        params['func'] = 'vss_del_snapshot'
        params['vssID'] = vssID
        params['snapshot_lun_NAA'] = snapLunNAA
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/snapshot.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

    @_connection_checker
    def cloneSnapshot(self, snapshot_id, new_lunname):
        """Execute CGI to clone snapshot as unmap lun."""
        params = {}
        params['func'] = 'clone_qsnapshot'
        params['by_lun'] = '1'
        params['snapshotID'] = snapshot_id
        params['new_name'] = new_lunname
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/snapshot.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

    @_connection_checker
    def editLun(self, lun):
        LOG.debug(
            _LI('LUNName:%(name)s, LUNCapacity:%(cap)s, LUNIndex:%(id)s'),
            ({'name': lun['LUNName'],
              'cap': lun['LUNCapacity'],
              'id': lun['LUNIndex']}))
        params = {}
        params['func'] = 'edit_lun'
        params['LUNName'] = lun['LUNName']
        params['LUNCapacity'] = lun['LUNCapacity']
        params['LUNIndex'] = lun['LUNIndex']
        params['LUNThinAllocate'] = lun['LUNThinAllocate']
        params['sid'] = self.sid
        params['LUNPath'] = lun['LUNPath']
        params['LUNStatus'] = lun['LUNStatus']

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_lun_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

    def addSnapsyncHost(self, ip, port, passwd):
        LOG.debug(_LI('ip: %(ip)s, port: %(port)s, passwd: %(passwd)s'),
                  {'ip': ip, 'port': port, 'passwd': passwd})
        params = {}
        params['func'] = 'host_add'
        params['ip'] = ip
        params['port'] = port
        params['passwd'] = passwd
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/snap_sync.cgi?%s' % sanitized_params)

        self._execute_and_get_response_details(self.ip, url)

    @_connection_checker
    def getHostidListByInitIQN(self, initIqn):
        """Execute getHostidListByInitIQN API."""
        LOG.debug(_LI('in getHostidListByInitIQN'))

        params = {}
        params['func'] = 'get_hostlist'
        params['module'] = 'hosts'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/accessrights/accessrightsRequest.cgi?%s' %
               sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        hostid = []
        hostname = []
        root = ET.fromstring(res_details['data'])
        hostList = root.find('content').find('host_list')
        hostListTree = hostList.findall('host')
        for host in hostListTree:
            LOG.debug(_LI('Iterating host name: %(name)s, hostid: %(id)s'),
                      {'name': host.find('name').text,
                      'id': host.find('hostid').text})
            iqns = host.find('iqns').findall('iqn')
            for iqn in iqns:
                # LOG.debug('iqn = %s ', iqn.text)
                if (initIqn == iqn.text):
                    hostid.append(host.find('hostid').text)
                    hostname.append(host.find('name').text)

        LOG.debug(_LI('hostid = %s'), hostid)
        LOG.debug(_LI('hostname = %s'), hostname)
        return hostid, hostname

    @_connection_checker
    def getAllIscsiPortalSetting(self):
        """Execute getAllIscsiPortalSetting API."""
        LOG.debug(_LI('in getAllIscsiPortalSetting'))

        params = {}
        params['func'] = 'get_all'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_portal_setting.cgi?%s' % sanitized_params)

        return self._execute_and_get_response_details(self.ip, url)

    @_connection_checker
    def getServiceBinding(self, **kwargs):
        """Execute getServiceBinding API."""
        LOG.debug(_LI('in getServiceBinding'))

        params = {}
        params['subfunc'] = 'service_binding'
        params['act'] = 'load_cfg'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/sys/servicebindingrequest.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        return res_details['data']

    @_connection_checker
    def getAllEthernetIp(self, **kwargs):
        """Execute getAllEthernetIp API."""
        LOG.debug(_LI('in getAllEthernetIp'))

        params = {}
        params['subfunc'] = 'net_setting'
        params['action'] = 'get_all_eth_ip'
        params['sid'] = self.sid
        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/sys/sysRequest.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        if ('type' in kwargs):
            returnIP = []
            IpList = root.find('func').find('ownContent')
            IpListTree = IpList.findall('IPConfig')
            for IP in IpListTree:
                LOG.debug(_LI('Iterating eth interface:%(eth)s, IP:%(ip)s'),
                          {'eth': IP.find('ethdev').text,
                          'ip': IP.find('IP').text})
                if ((kwargs['type'] == 'data') and
                   (IP.find('ethdev').text != 'em0') and
                   (IP.find('IP').text is not None)):
                    returnIP.append(IP.find('IP').text)
                elif ((kwargs['type'] == 'manage') and
                      (IP.find('ethdev').text == 'em0') and
                      (IP.find('IP').text is not None)):
                    returnIP.append(IP.find('IP').text)
                elif (kwargs['type'] == 'all'):
                    dict = {}
                    for child in IP:
                        dict[child.tag] = child.text
                    returnIP.append(dict)
            LOG.debug(_LI('returnIP = %s'), returnIP)
            return returnIP
        else:
            return res_details['data']


class QnapAPIExecutorTS(QnapAPIExecutor):
    """Makes QNAP API calls for TS NAS."""

    @_connection_checker
    def getPoolInfo(self, **kwargs):
        """Execute deleteInitiatorGrp API."""
        for key, value in six.iteritems(kwargs):
            LOG.debug(_LI('%(key)s = %(val)s'),
                      {'key': key, 'val': value})

        params = {}
        params['Pool_Info'] = '1'
        params['func'] = 'extra_get'
        params['poolID'] = '1'
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/disk_manage.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')

        if 'poolIndex' in kwargs:
            LOG.debug(_LI('in poolIndex section...'))
            root = ET.fromstring(res_details['data'])
            poolList = root.find('Pool_Index')
            poolInfoTree = poolList.findall('row')
            for pool in poolInfoTree:
                LOG.debug(_LI('Iterating pool name: %(name)s, index: %(id)s'),
                          {'name': pool.find('poolID').text,
                          'id': pool.find('poolIndex').text})
                if kwargs['poolIndex'] == pool.find('poolIndex').text:
                    LOG.debug(_LI('poolIndex: %s'),
                              pool.find('poolIndex').text)
                    return pool
        elif 'poolID' in kwargs:
            LOG.debug(_LI('in poolID section...'))
            root = ET.fromstring(res_details['data'])
            poolList = root.find('Pool_Index')
            poolInfoTree = poolList.findall('row')
            for pool in poolInfoTree:
                if kwargs['poolID'] == pool.find('poolID').text:
                    LOG.debug(_LI('poolID: %s'), pool.find('poolID').text)
                    return pool
        else:
            return res_details['data']

    @_connection_checker
    def removeTargetInit(self, targetIqn, initIqn):
        LOG.debug(_LI('targetIqn = %(tgt)s, initIqn = %(init)s'),
                  {'tgt': targetIqn, 'init': initIqn})

        params = {}
        params['func'] = 'remove_init'
        params['targetIQN'] = targetIqn
        params['initiatorIQN'] = initIqn
        params['ha_sync'] = '1'
        params['sid'] = self.sid

        sanitized_params = {}

        for key in params:
            value = params[key]
            if value is not None:
                sanitized_params[key] = six.text_type(value)

        sanitized_params = urllib.parse.urlencode(sanitized_params)
        url = ('/cgi-bin/disk/iscsi_target_setting.cgi?%s' % sanitized_params)

        res_details = self._execute_and_get_response_details(self.ip, url)
        root = ET.fromstring(res_details['data'])
        if root.find('authPassed').text == '0':
            raise exception.QnapAPIExpiredException('session id expired')
